//-----------------------------------------------------------------------------
// Copyright (c) 2008 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Globals.
//-----------------------------------------------------------------------------

float4x4 world;
float4x4 worldInvTrans;
float4x4 worldViewProjection;

float3 sunlightDir;
float4 sunlightColor;
float4 terrainAmbient;
float4 terrainDiffuse;
float terrainTilingFactor;

float2 terrainRegion1;
float2 terrainRegion2;
float2 terrainRegion3;
float2 terrainRegion4;

//-----------------------------------------------------------------------------
// Textures.
//-----------------------------------------------------------------------------

texture region1ColorMapTexture;
sampler2D region1ColorMap = sampler_state
{
	Texture = <region1ColorMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
    AddressU = WRAP;
    AddressV = WRAP;
};

texture region1NormalMapTexture;
sampler2D region1NormalMap = sampler_state
{
	Texture = <region1NormalMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
    AddressU = WRAP;
    AddressV = WRAP;
};

texture region2ColorMapTexture;
sampler2D region2ColorMap = sampler_state
{
	Texture = <region2ColorMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
    AddressU = WRAP;
    AddressV = WRAP;
};

texture region2NormalMapTexture;
sampler2D region2NormalMap = sampler_state
{
	Texture = <region2NormalMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
    AddressU = WRAP;
    AddressV = WRAP;
};

texture region3ColorMapTexture;
sampler2D region3ColorMap = sampler_state
{
	Texture = <region3ColorMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
    AddressU = WRAP;
    AddressV = WRAP;
};

texture region3NormalMapTexture;
sampler2D region3NormalMap = sampler_state
{
	Texture = <region3NormalMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
    AddressU = WRAP;
    AddressV = WRAP;
};

texture region4ColorMapTexture;
sampler2D region4ColorMap = sampler_state
{
	Texture = <region4ColorMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
    AddressU = WRAP;
    AddressV = WRAP;
};

texture region4NormalMapTexture;
sampler2D region4NormalMap = sampler_state
{
	Texture = <region4NormalMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
    AddressU = WRAP;
    AddressV = WRAP;
};

//-----------------------------------------------------------------------------
// Vertex Shaders.
//-----------------------------------------------------------------------------

void VS_Main(in  float3 inPos       : POSITION,
			 in  float2 inTexCoord  : TEXCOORD,
			 in  float3 inNormal    : NORMAL,
			 in  float3 inTangent   : TANGENT,
			 out float4 outPos      : POSITION,
			 out float2 outTexCoord : TEXCOORD0,
			 out float4 outLightDir : TEXCOORD1)
{
    float3 n = mul(inNormal, (float3x3)worldInvTrans);
	float3 t = mul(inTangent, (float3x3)worldInvTrans);
	float3 b = cross(n, t);
	float3x3 tbnMatrix = float3x3(t.x, b.x, n.x,
	                              t.y, b.y, n.y,
	                              t.z, b.z, n.z);
    
	outPos = mul(float4(inPos, 1.0f), worldViewProjection);
	outTexCoord = inTexCoord * terrainTilingFactor;
	
	outLightDir.xyz = -sunlightDir;
	outLightDir.xyz = mul(outLightDir.xyz, tbnMatrix);
			
	// Store the terrain height at this vertex position.
	// The terrain height will be used to generate the terrain texture.
	outLightDir.w = inPos.y;
}

//-----------------------------------------------------------------------------
// Pixel Shaders.
//-----------------------------------------------------------------------------

float4 PS_GetTerrainColor(const sampler2D colorMap,
                          const sampler2D normalMap,
                          const float2 texCoord,
                          const float3 lightDir)
{
    float3 n = normalize(tex2D(normalMap, texCoord).rgb * 2.0f - 1.0f);   
    float4 ambient = sunlightColor * terrainAmbient;
    float4 diffuse = sunlightColor * terrainDiffuse * saturate(dot(n, lightDir));
	               
	return (ambient + diffuse) * tex2D(colorMap, texCoord);
}

void PS_Main(in  float2 inTexCoord : TEXCOORD0,
		     in  float4 inLightDir : TEXCOORD1,
		     out float4 outColor   : COLOR)
{	
	float3 lightDir = normalize(inLightDir.xyz);
	float height = inLightDir.w;
	
	float regionMin;
	float regionMax;
	float regionRange;
	float regionWeight;
	
	outColor = float4(0.0f, 0.0f, 0.0f, 0.0f);
		
	// Terrain region 1.
	
    regionMin = terrainRegion1.x;
    regionMax = terrainRegion1.y;
    regionRange = regionMax - regionMin;
    regionWeight = (regionRange - abs(height - regionMax)) / regionRange;
    regionWeight = saturate(regionWeight);
    outColor += regionWeight * PS_GetTerrainColor(region1ColorMap, region1NormalMap, inTexCoord, lightDir);
			
	// Terrain region 2.
	
	regionMin = terrainRegion2.x;
    regionMax = terrainRegion2.y;
    regionRange = regionMax - regionMin;
    regionWeight = (regionRange - abs(height - regionMax)) / regionRange;
    regionWeight = saturate(regionWeight);
    outColor += regionWeight * PS_GetTerrainColor(region2ColorMap, region2NormalMap, inTexCoord, lightDir);
			
	// Terrain region 3.
	
	regionMin = terrainRegion3.x;
    regionMax = terrainRegion3.y;
    regionRange = regionMax - regionMin;
    regionWeight = (regionRange - abs(height - regionMax)) / regionRange;
    regionWeight = saturate(regionWeight);
    outColor += regionWeight * PS_GetTerrainColor(region3ColorMap, region3NormalMap, inTexCoord, lightDir);
	
	// Terrain region 4.
	
	regionMin = terrainRegion4.x;
    regionMax = terrainRegion4.y;
    regionRange = regionMax - regionMin;
    regionWeight = (regionRange - abs(height - regionMax)) / regionRange;
    regionWeight = saturate(regionWeight);
    outColor += regionWeight * PS_GetTerrainColor(region4ColorMap, region4NormalMap, inTexCoord, lightDir);
}

//-----------------------------------------------------------------------------
// Techniques.
//-----------------------------------------------------------------------------

technique TerrainTexturing
{
    pass
    {
        VertexShader = compile vs_2_0 VS_Main();
        PixelShader = compile ps_2_0 PS_Main();
    }
}
